#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Literal, Optional, Sequence, Tuple, Union, overload

import torch
from pythonwrench.semver import Version
from pythonwrench.typing import BuiltinNumber
from torch import Tensor
from torch.types import Number
from typing_extensions import Never

from torchwrench.core.dtype_enum import DTypeEnum
from torchwrench.core.make import (
    DeviceLike,
    DTypeLike,
    GeneratorLike,
    as_device,
    as_dtype,
    as_generator,
)
from torchwrench.types import (
    BoolTensor0D,
    BoolTensor1D,
    BoolTensor2D,
    BoolTensor3D,
    CFloatTensor0D,
    CFloatTensor1D,
    CFloatTensor2D,
    CFloatTensor3D,
    FloatTensor0D,
    FloatTensor1D,
    FloatTensor2D,
    FloatTensor3D,
    LongTensor0D,
    LongTensor1D,
    LongTensor2D,
    LongTensor3D,
    Tensor0D,
    Tensor1D,
    Tensor2D,
    Tensor3D,
)

__all__ = [
    "arange",
    "empty",
    "full",
    "rand",
    "randint",
    "randperm",
    "ones",
    "zeros",
]

# ----------
# arange
# ----------


@overload
def arange(
    end: Number,
    *,
    out: Optional[Tensor] = None,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> LongTensor1D: ...


@overload
def arange(
    start: Number,
    end: Number,
    *,
    out: Optional[Tensor] = None,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> LongTensor1D: ...


@overload
def arange(
    start: Number,
    end: Number,
    step: Number,
    *,
    out: Optional[Tensor] = None,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> LongTensor1D: ...


@overload
def arange(
    end: Number,
    *,
    out: Optional[Tensor] = None,
    dtype: Union[torch.dtype, str, DTypeEnum],
    device: DeviceLike = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> Tensor1D: ...


@overload
def arange(
    start: Number,
    end: Number,
    *,
    out: Optional[Tensor] = None,
    dtype: Union[torch.dtype, str, DTypeEnum],
    device: DeviceLike = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> Tensor1D: ...


@overload
def arange(
    start: Number,
    end: Number,
    step: Number,
    *,
    out: Optional[Tensor] = None,
    dtype: Union[torch.dtype, str, DTypeEnum],
    device: DeviceLike = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> Tensor1D: ...


def arange(
    *args: Number,
    out: Optional[Tensor] = None,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
    **kwargs,
) -> torch.Tensor:
    dtype = as_dtype(dtype)
    device = as_device(device)

    kwds = {}
    if Version(torch.__version__) >= Version("2.0.0"):
        kwds.update(pin_memory=pin_memory)

    return torch.arange(
        *args,
        **kwargs,
        out=out,
        dtype=dtype,
        device=device,
        requires_grad=requires_grad,
        **kwds,
    )


# ----------
# empty
# ----------


@overload
def empty(
    size: Sequence[Never],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    memory_format: Optional[torch.memory_format] = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor0D: ...


@overload
def empty(
    size: Tuple[int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    memory_format: Optional[torch.memory_format] = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor1D: ...


@overload
def empty(
    size: Tuple[int, int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    memory_format: Optional[torch.memory_format] = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor2D: ...


@overload
def empty(
    size: Tuple[int, int, int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    memory_format: Optional[torch.memory_format] = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor3D: ...


@overload
def empty(
    size0: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    memory_format: Optional[torch.memory_format] = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor1D: ...


@overload
def empty(
    size0: int,
    size1: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    memory_format: Optional[torch.memory_format] = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor2D: ...


@overload
def empty(
    size0: int,
    size1: int,
    size2: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    memory_format: Optional[torch.memory_format] = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor3D: ...


def empty(
    *data,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    memory_format: Optional[torch.memory_format] = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> torch.Tensor:
    dtype = as_dtype(dtype)
    device = as_device(device)

    if layout is None:
        layout = torch.strided
    if pin_memory is None:
        pin_memory = False
    if requires_grad is None:
        requires_grad = False

    return torch.empty(  # type: ignore
        *data,
        memory_format=memory_format,
        out=out,
        dtype=dtype,  # type: ignore
        layout=layout,
        device=device,
        pin_memory=pin_memory,
        requires_grad=requires_grad,
    )


# ----------
# full
# ----------


# bool
@overload
def full(  # type: ignore
    size: Sequence[Never],
    fill_value: bool,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor0D: ...


@overload
def full(  # type: ignore
    size: Tuple[int],
    fill_value: bool,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor1D: ...


@overload
def full(  # type: ignore
    size: Tuple[int, int],
    fill_value: bool,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor2D: ...


@overload
def full(  # type: ignore
    size: Tuple[int, int, int],
    fill_value: bool,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> BoolTensor3D: ...


# int
@overload
def full(
    size: Sequence[Never],
    fill_value: int,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> LongTensor0D: ...


@overload
def full(
    size: Tuple[int],
    fill_value: int,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> LongTensor1D: ...


@overload
def full(
    size: Tuple[int, int],
    fill_value: int,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> LongTensor2D: ...


@overload
def full(
    size: Tuple[int, int, int],
    fill_value: int,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> LongTensor3D: ...


# float
@overload
def full(
    size: Sequence[Never],
    fill_value: float,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> FloatTensor0D: ...


@overload
def full(
    size: Tuple[int],
    fill_value: float,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> FloatTensor1D: ...


@overload
def full(
    size: Tuple[int, int],
    fill_value: float,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> FloatTensor2D: ...


@overload
def full(
    size: Tuple[int, int, int],
    fill_value: float,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> FloatTensor3D: ...


# complex
@overload
def full(
    size: Sequence[Never],
    fill_value: complex,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> CFloatTensor0D: ...


@overload
def full(
    size: Tuple[int],
    fill_value: complex,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> CFloatTensor1D: ...


@overload
def full(
    size: Tuple[int, int],
    fill_value: complex,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> CFloatTensor2D: ...


@overload
def full(
    size: Tuple[int, int, int],
    fill_value: complex,
    *,
    dtype: Literal[None] = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> CFloatTensor3D: ...


# BuiltinNumber
@overload
def full(
    size: Sequence[Never],
    fill_value: BuiltinNumber,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor0D: ...


@overload
def full(
    size: Tuple[int],
    fill_value: BuiltinNumber,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor1D: ...


@overload
def full(
    size: Tuple[int, int],
    fill_value: BuiltinNumber,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor2D: ...


@overload
def full(
    size: Tuple[int, int, int],
    fill_value: BuiltinNumber,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor3D: ...


def full(
    size: Sequence[int],
    fill_value: BuiltinNumber,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> torch.Tensor:
    dtype = as_dtype(dtype)
    device = as_device(device)
    kwds = {}
    if Version(torch.__version__) >= Version("2.0.0"):
        kwds.update(pin_memory=pin_memory)
    if layout is None:
        layout = torch.strided
    if requires_grad is None:
        requires_grad = False

    return torch.full(
        tuple(size),
        fill_value,  # type: ignore
        out=out,
        dtype=dtype,
        layout=layout,
        device=device,
        requires_grad=requires_grad,
        **kwds,
    )


# ----------
# ones
# ----------


@overload
def ones(
    size: Sequence[Never],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor0D: ...


@overload
def ones(
    size: Tuple[int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor1D: ...


@overload
def ones(
    size: Tuple[int, int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor2D: ...


@overload
def ones(
    size: Tuple[int, int, int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor3D: ...


@overload
def ones(
    size0: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor1D: ...


@overload
def ones(
    size0: int,
    size1: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor2D: ...


@overload
def ones(
    size0: int,
    size1: int,
    size2: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor3D: ...


def ones(
    *data,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> torch.Tensor:
    dtype = as_dtype(dtype)
    device = as_device(device)

    if layout is None:
        layout = torch.strided
    if pin_memory is None:
        pin_memory = False
    if requires_grad is None:
        requires_grad = False

    return torch.ones(  # type: ignore
        *data,
        out=out,
        dtype=dtype,  # type: ignore
        layout=layout,
        device=device,
        pin_memory=pin_memory,
        requires_grad=requires_grad,
    )


# ----------
# rand
# ----------
@overload
def rand(
    size: Sequence[Never],
    /,
    *,
    dtype: Literal["float", "float32", None] = None,
    device: DeviceLike = None,
    generator: GeneratorLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> FloatTensor0D: ...


@overload
def rand(
    size: Tuple[int],
    /,
    *,
    dtype: Literal["float", "float32", None] = None,
    device: DeviceLike = None,
    generator: GeneratorLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> FloatTensor1D: ...


@overload
def rand(
    size: Tuple[int, int],
    /,
    *,
    dtype: Literal["float", "float32", None] = None,
    device: DeviceLike = None,
    generator: GeneratorLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> FloatTensor2D: ...


@overload
def rand(
    size: Tuple[int, int, int],
    /,
    *,
    dtype: Literal["float", "float32", None] = None,
    device: DeviceLike = None,
    generator: GeneratorLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> FloatTensor3D: ...


@overload
def rand(
    size0: int,
    /,
    *,
    dtype: Literal["float", "float32", None] = None,
    device: DeviceLike = None,
    generator: GeneratorLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> FloatTensor1D: ...


@overload
def rand(
    size0: int,
    size1: int,
    /,
    *,
    dtype: Literal["float", "float32", None] = None,
    device: DeviceLike = None,
    generator: GeneratorLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> FloatTensor2D: ...


@overload
def rand(
    size0: int,
    size1: int,
    size2: int,
    /,
    *,
    dtype: Literal["float", "float32", None] = None,
    device: DeviceLike = None,
    generator: GeneratorLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> FloatTensor3D: ...


@overload
def rand(
    size: Sequence[Never],
    /,
    *,
    dtype: DTypeLike,
    device: DeviceLike = None,
    generator: GeneratorLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor0D: ...


@overload
def rand(
    size: Tuple[int],
    /,
    *,
    dtype: DTypeLike,
    device: DeviceLike = None,
    generator: GeneratorLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor1D: ...


@overload
def rand(
    size: Tuple[int, int],
    /,
    *,
    dtype: DTypeLike,
    device: DeviceLike = None,
    generator: GeneratorLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor2D: ...


@overload
def rand(
    size: Tuple[int, int, int],
    /,
    *,
    dtype: DTypeLike,
    device: DeviceLike = None,
    generator: GeneratorLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor3D: ...


@overload
def rand(
    size0: int,
    /,
    *,
    dtype: DTypeLike,
    device: DeviceLike = None,
    generator: GeneratorLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor1D: ...


@overload
def rand(
    size0: int,
    size1: int,
    /,
    *,
    dtype: DTypeLike,
    device: DeviceLike = None,
    generator: GeneratorLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor2D: ...


@overload
def rand(
    size0: int,
    size1: int,
    size2: int,
    /,
    *,
    dtype: DTypeLike,
    device: DeviceLike = None,
    generator: GeneratorLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor3D: ...


def rand(
    *data: Any,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    generator: GeneratorLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> torch.Tensor:
    dtype = as_dtype(dtype)
    device = as_device(device)
    generator = as_generator(generator)

    kwds = {}
    if Version(torch.__version__) >= Version("2.0.0"):
        kwds.update(pin_memory=pin_memory)
    if layout is None:
        layout = torch.strided
    if requires_grad is None:
        requires_grad = False

    return torch.rand(  # type: ignore
        *data,
        generator=generator,
        out=out,
        dtype=dtype,  # type: ignore
        layout=layout,
        device=device,
        requires_grad=requires_grad,
        **kwds,
    )


# ----------
# randint
# ----------


@overload
def randint(
    low: int,
    high: int,
    size: Tuple[()],
    *,
    generator: GeneratorLike = None,
    dtype: Literal[None, "long", "int64"] = None,
    device: DeviceLike = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> LongTensor0D: ...


@overload
def randint(
    low: int,
    high: int,
    size: Tuple[int],
    *,
    generator: GeneratorLike = None,
    dtype: Literal[None, "long", "int64"] = None,
    device: DeviceLike = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> LongTensor1D: ...


@overload
def randint(
    low: int,
    high: int,
    size: Tuple[int, int],
    *,
    generator: GeneratorLike = None,
    dtype: Literal[None, "long", "int64"] = None,
    device: DeviceLike = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> LongTensor2D: ...


@overload
def randint(
    low: int,
    high: int,
    size: Tuple[int, int, int],
    *,
    generator: GeneratorLike = None,
    dtype: Literal[None, "long", "int64"] = None,
    device: DeviceLike = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> LongTensor3D: ...


@overload
def randint(
    low: int,
    high: int,
    size: Tuple[()],
    *,
    generator: GeneratorLike = None,
    dtype: Union[torch.dtype, str, DTypeLike],
    device: DeviceLike = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> Tensor0D: ...


@overload
def randint(
    low: int,
    high: int,
    size: Tuple[int],
    *,
    generator: GeneratorLike = None,
    dtype: Union[torch.dtype, str, DTypeLike],
    device: DeviceLike = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> Tensor1D: ...


@overload
def randint(
    low: int,
    high: int,
    size: Tuple[int, int],
    *,
    generator: GeneratorLike = None,
    dtype: Union[torch.dtype, str, DTypeLike],
    device: DeviceLike = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> Tensor2D: ...


@overload
def randint(
    low: int,
    high: int,
    size: Tuple[int, int, int],
    *,
    generator: GeneratorLike = None,
    dtype: Union[torch.dtype, str, DTypeLike],
    device: DeviceLike = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> Tensor3D: ...


def randint(
    low: int,
    high: int,
    size: Tuple[int, ...],
    *,
    generator: GeneratorLike = None,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    requires_grad: bool = False,
    pin_memory: bool = False,
) -> Tensor:
    dtype = as_dtype(dtype)
    device = as_device(device)
    generator = as_generator(generator)

    kwds = {}
    if Version(torch.__version__) >= Version("2.0.0"):
        kwds.update(pin_memory=pin_memory)

    return torch.randint(
        low=low,
        high=high,
        size=size,
        generator=generator,
        dtype=dtype,
        device=device,
        requires_grad=requires_grad,
        **kwds,
    )


# ----------
# randperm
# ----------


@overload
def randperm(
    n: int,
    *,
    generator: GeneratorLike = None,
    out: Optional[Tensor] = None,
    dtype: Literal[None, "long", "int64"] = "long",
    layout: Optional[torch.layout] = None,
    device: DeviceLike = None,
    pin_memory: Optional[bool] = False,
    requires_grad: Optional[bool] = False,
) -> LongTensor1D: ...


@overload
def randperm(
    n: int,
    *,
    generator: GeneratorLike = None,
    out: Optional[Tensor] = None,
    dtype: Union[torch.dtype, str, DTypeLike],
    layout: Optional[torch.layout] = None,
    device: DeviceLike = None,
    pin_memory: Optional[bool] = False,
    requires_grad: Optional[bool] = False,
) -> Tensor1D: ...


def randperm(
    n: int,
    *,
    generator: GeneratorLike = None,
    out: Optional[Tensor] = None,
    dtype: DTypeLike = "long",
    layout: Optional[torch.layout] = None,
    device: DeviceLike = None,
    pin_memory: Optional[bool] = False,
    requires_grad: Optional[bool] = False,
) -> Tensor:
    dtype = as_dtype(dtype)
    device = as_device(device)
    generator = as_generator(generator)

    if layout is None:
        layout = torch.strided
    if pin_memory is None:
        pin_memory = False
    if requires_grad is None:
        requires_grad = False

    return torch.randperm(
        n=n,
        generator=generator,
        out=out,
        dtype=dtype,  # type: ignore
        layout=layout,
        device=device,
        pin_memory=pin_memory,
        requires_grad=requires_grad,
    )


# ----------
# zeros
# ----------
@overload
def zeros(
    size: Sequence[Never],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor0D: ...


@overload
def zeros(
    size: Tuple[int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor1D: ...


@overload
def zeros(
    size: Tuple[int, int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor2D: ...


@overload
def zeros(
    size: Tuple[int, int, int],
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor3D: ...


@overload
def zeros(
    size0: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor1D: ...


@overload
def zeros(
    size0: int,
    size1: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor2D: ...


@overload
def zeros(
    size0: int,
    size1: int,
    size2: int,
    /,
    *,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> Tensor3D: ...


def zeros(
    *data,
    dtype: DTypeLike = None,
    device: DeviceLike = None,
    out: Union[torch.Tensor, None] = None,
    layout: Union[torch.layout, None] = None,
    pin_memory: Union[bool, None] = False,
    requires_grad: Union[bool, None] = False,
) -> torch.Tensor:
    dtype = as_dtype(dtype)
    device = as_device(device)

    if layout is None:
        layout = torch.strided
    if pin_memory is None:
        pin_memory = False
    if requires_grad is None:
        requires_grad = False

    return torch.zeros(  # type: ignore
        *data,
        out=out,
        dtype=dtype,  # type: ignore
        layout=layout,
        device=device,
        pin_memory=pin_memory,
        requires_grad=requires_grad,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
from typing import Any, Generic, Iterable, List, Tuple, TypeVar, Union, overload

import pythonwrench as pw
import torch
from pythonwrench.typing.classes import SupportsGetitemLen
from torch.utils.data.dataset import Dataset
from typing_extensions import TypeAlias

from torchwrench.extras.numpy import np
from torchwrench.extras.numpy.functional import is_numpy_bool_array
from torchwrench.nn.functional.transform import as_tensor
from torchwrench.types._typing import BoolTensor1D, Tensor1D, TensorOrArray
from torchwrench.types.guards import is_number_like, is_tensor_or_array
from torchwrench.utils.data.dataset.wrapper import Wrapper

T = TypeVar("T", covariant=True)
U = TypeVar("U", covariant=True)

MultiIndexer: TypeAlias = Union[
    Iterable[bool],
    Iterable[int],
    None,
    slice,
    Tensor1D,
    np.ndarray,
]


class DatasetSlicer(Generic[T], ABC, Dataset[T]):
    def __init__(
        self,
        *,
        add_slice_support: bool = True,
        add_indices_support: bool = True,
        add_mask_support: bool = True,
        add_none_support: bool = True,
    ) -> None:
        Dataset.__init__(self)
        self._add_slice_support = add_slice_support
        self._add_indices_support = add_indices_support
        self._add_mask_support = add_mask_support
        self._add_none_support = add_none_support

    @abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_item(self, idx, /, *args, **kwargs) -> Any:
        raise NotImplementedError

    @overload
    def __getitem__(self, idx: int, /) -> T:  # type: ignore
        ...

    @overload
    def __getitem__(self, idx: MultiIndexer, /) -> List[T]:  # type: ignore
        ...

    @overload
    def __getitem__(self, idx: Tuple[Any, ...], /) -> Any:  # type: ignore
        ...

    def __getitem__(self, idx) -> Any:
        if isinstance(idx, tuple) and len(idx) > 1:
            idx, *args = idx
        else:
            args = ()

        if is_number_like(idx):
            return self.get_item(idx, *args)

        elif isinstance(idx, slice):
            return self.get_items_slice(idx, *args)

        elif (
            pw.isinstance_generic(idx, Iterable[bool])
            or isinstance(idx, BoolTensor1D)
            or (is_numpy_bool_array(idx) and idx.ndim == 1)
        ):
            return self.get_items_mask(idx, *args)

        elif pw.isinstance_generic(idx, Iterable[int]) or is_tensor_or_array(idx):
            return self.get_items_indices(idx, *args)

        elif idx is None:
            return self.get_items_none(idx, *args)

        else:
            raise TypeError(f"Invalid argument type {type(idx)=} with {args=}.")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def get_items_indices(
        self,
        indices: Union[Iterable[int], TensorOrArray],
        *args,
    ) -> List[T]:
        if self._add_indices_support:
            return [self.get_item(idx, *args) for idx in indices]
        else:
            return self.get_item(indices, *args)

    def get_items_mask(
        self,
        mask: Union[Iterable[bool], TensorOrArray],
        *args,
    ) -> List[T]:
        if self._add_mask_support:
            mask = as_tensor(mask, dtype=torch.bool)
            if len(mask) > 0 and len(mask) != len(self):  # type: ignore
                msg = f"Invalid mask size {len(mask)}. (expected {len(self)})"
                raise ValueError(msg)

            indices = torch.where(mask)[0]
            return self.get_items_indices(indices, *args)
        else:
            return self.get_item(mask, *args)

    def get_items_slice(
        self,
        slice_: slice,
        *args,
    ) -> List[T]:
        if self._add_slice_support:
            return self.get_items_indices(range(len(self))[slice_], *args)
        else:
            return self.get_item(slice_, *args)

    def get_items_none(
        self,
        none: None,
        *args,
    ) -> List[T]:
        if self._add_none_support:
            return self.get_items_slice(slice(None), *args)
        else:
            return self.get_item(none, *args)


class DatasetSlicerWrapper(Generic[T], DatasetSlicer[T], Wrapper[T]):
    def __init__(
        self,
        dataset: SupportsGetitemLen[T],
        *,
        add_slice_support: bool = True,
        add_indices_support: bool = True,
        add_mask_support: bool = True,
        add_none_support: bool = True,
    ) -> None:
        """Wrap a sequence to support slice, indices and mask arguments types."""
        DatasetSlicer.__init__(
            self,
            add_slice_support=add_slice_support,
            add_indices_support=add_indices_support,
            add_mask_support=add_mask_support,
            add_none_support=add_none_support,
        )
        Wrapper.__init__(self, dataset)

    def __len__(self) -> int:
        return len(self.dataset)

    def get_item(self, idx: int, *args) -> T:
        # note: we need to split calls here, because self.dataset[idx] give an int as argument while self.dataset[idx, *args] always gives a tuple even if args == ()
        if len(args) == 0:
            return self.dataset[idx]
        else:
            # equivalent to self.dataset[idx, *args], but only in recent python versions
            return self.dataset.__getitem__((idx,) + args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
from argparse import Namespace
from pathlib import Path
from typing import Any, Iterable, Literal, Mapping, Optional, Type, Union

from pythonwrench.functools import function_alias
from pythonwrench.typing import DataclassInstance, NamedTupleInstance
from typing_extensions import TypeAlias

from torchwrench.core.packaging import _OMEGACONF_AVAILABLE, _YAML_AVAILABLE
from torchwrench.serialization.common import as_builtin

if not _YAML_AVAILABLE:
    from torchwrench.extras.yaml import _yaml_fallback as yaml
    from torchwrench.extras.yaml._yaml_fallback import (
        BaseLoader,
        CBaseLoader,
        CFullLoader,
        CLoader,
        CSafeLoader,
        CUnsafeLoader,
        FullLoader,
        Loader,
        MappingNode,
        Node,
        ParserError,
        SafeLoader,
        ScalarNode,
        ScannerError,
        SequenceNode,
        UnsafeLoader,
    )

else:
    import yaml
    from yaml import (
        BaseLoader,
        FullLoader,
        Loader,
        MappingNode,
        Node,
        SafeLoader,
        ScalarNode,
        SequenceNode,
        UnsafeLoader,
    )
    from yaml.parser import ParserError
    from yaml.scanner import ScannerError

    try:
        from yaml import (
            CBaseLoader,
            CFullLoader,
            CLoader,
            CSafeLoader,
            CUnsafeLoader,
        )
    except ImportError:
        from torchwrench.extras.yaml._yaml_fallback import (
            CBaseLoader,
            CFullLoader,
            CLoader,
            CSafeLoader,
            CUnsafeLoader,
        )


if _OMEGACONF_AVAILABLE:
    from omegaconf import OmegaConf  # type: ignore


YamlLoaders: TypeAlias = Union[
    Type[Loader],
    Type[BaseLoader],
    Type[FullLoader],
    Type[SafeLoader],
    Type[UnsafeLoader],
    Type[CLoader],
    Type[CBaseLoader],
    Type[CFullLoader],
    Type[CSafeLoader],
    Type[CUnsafeLoader],
]


def dump_yaml(
    data: Union[
        Iterable[Any],
        Mapping[str, Any],
        Namespace,
        DataclassInstance,
        NamedTupleInstance,
    ],
    fpath: Union[str, Path, None] = None,
    *,
    overwrite: bool = True,
    to_builtins: bool = False,
    make_parents: bool = True,
    resolve: bool = False,
    encoding: Optional[str] = "utf-8",
    # YAML dump kwargs
    sort_keys: bool = False,
    indent: Union[int, None] = None,
    width: Union[int, None] = 1000,
    allow_unicode: bool = True,
    **yaml_dump_kwds,
) -> str:
    """Dump content to yaml format."""
    if not _YAML_AVAILABLE:
        msg = f"Cannot use python module {__file__} since pyyaml package is not installed. Please install it with `pip install torchwrench[extras]`."
        raise ImportError(msg)

    if not _OMEGACONF_AVAILABLE and resolve:
        msg = (
            "Cannot resolve yaml config without omegaconf package."
            "Please use resolve=False or install omegaconf with `pip install torchwrench[extras]`."
        )
        raise ValueError(msg)

    if fpath is not None:
        fpath = Path(fpath).resolve().expanduser()
        if not overwrite and fpath.exists():
            raise FileExistsError(f"File {fpath} already exists.")
        elif make_parents:
            fpath.parent.mkdir(parents=True, exist_ok=True)

    if resolve:
        data = OmegaConf.create(data)  # type: ignore
        data = OmegaConf.to_container(data, resolve=True)  # type: ignore

    if to_builtins:
        data = as_builtin(data)

    content = yaml.dump(
        data,
        sort_keys=sort_keys,
        indent=indent,
        width=width,
        allow_unicode=allow_unicode,
        **yaml_dump_kwds,
    )
    if fpath is not None:
        fpath.write_text(content, encoding=encoding)
    return content


@function_alias(dump_yaml)
def dumps_yaml(*args, **kwargs): ...


@function_alias(dump_yaml)
def save_yaml(*args, **kwargs): ...


def load_yaml(
    file: Union[str, Path, io.TextIOBase],
    *,
    Loader: YamlLoaders = SafeLoader,
    on_error: Literal["raise", "ignore"] = "raise",
) -> Any:
    """Load YAML from filepath or opened file."""
    if not _YAML_AVAILABLE:
        msg = f"Cannot use python module {__file__} since pyyaml package is not installed. Please install it with `pip install torchwrench[extras]`."
        raise ImportError(msg)

    if isinstance(file, (str, Path)):
        with open(file, "r") as buffer:
            return loads_yaml(buffer, Loader=Loader, on_error=on_error)
    elif isinstance(file, io.TextIOBase):
        return loads_yaml(file, Loader=Loader, on_error=on_error)
    else:
        msg = f"Invalid argument type {type(file)}."
        raise TypeError(msg)


def loads_yaml(
    content: Union[str, io.TextIOBase],
    *,
    Loader: YamlLoaders = SafeLoader,
    on_error: Literal["raise", "ignore"] = "raise",
) -> Any:
    """Load YAML from string and text-io stream."""
    if isinstance(content, str):
        with io.StringIO(content) as buffer:
            return loads_yaml(buffer, Loader=Loader, on_error=on_error)

    try:
        data = yaml.load(content, Loader=Loader)  # type: ignore
    except (ScannerError, ParserError) as err:
        if on_error == "ignore":
            return None
        else:
            raise err
    return data


@function_alias(load_yaml)
def read_yaml(*args, **kwargs): ...


class IgnoreTagLoader(SafeLoader):  # type: ignore
    """SafeLoader that ignores yaml tags.

    Examples
    ========

    ```python
    >>> dumped = "a: !!python/tuple\n- 1\n- 2"
    >>> yaml.load(dumped, Loader=IgnoreTagLoader)
    ... {"a": [1, 2]}
    >>> yaml.load(dumped, Loader=FullLoader)
    ... {"a": (1, 2)}
    >>> yaml.load(dumped, Loader=SafeLoader)  # raises ConstructorError
    ```
    """

    def construct_with_tag(self, tag: str, node: Node) -> Any:
        if isinstance(node, MappingNode):
            return self.construct_mapping(node)
        elif isinstance(node, ScalarNode):
            return self.construct_scalar(node)
        elif isinstance(node, SequenceNode):
            return self.construct_sequence(node)
        else:
            msg = f"Unsupported node type {type(node)} with {tag=}."
            raise NotImplementedError(msg)


class SplitTagLoader(SafeLoader):  # type: ignore
    """SafeLoader that store tags inside value.

    Examples
    ========

    ```python
    >>> dumped = "a: !!python/tuple\n- 1\n- 2"
    >>> yaml.load(dumped, Loader=SplitTagLoader)
    ... {'a': {'_target_': 'yaml.org,2002:python/tuple', '_args_': [1, 2]}}
    ```
    """

    def __init__(
        self,
        stream,
        *,
        tag_key: str = "_target_",
        args_key: str = "_args_",
    ) -> None:
        super().__init__(stream)
        self.tag_key = tag_key
        self.args_key = args_key

    def construct_with_tag(self, tag: str, node: Node) -> Any:
        if isinstance(node, MappingNode):
            result = self.construct_mapping(node)
        elif isinstance(node, ScalarNode):
            result = self.construct_scalar(node)
        elif isinstance(node, SequenceNode):
            result = self.construct_sequence(node)
        else:
            msg = f"Unsupported node type {type(node)} with {tag=}."
            raise NotImplementedError(msg)

        result = {
            self.tag_key: tag,
            self.args_key: result,
        }
        return result


if _YAML_AVAILABLE:
    IgnoreTagLoader.add_multi_constructor("!", IgnoreTagLoader.construct_with_tag)
    IgnoreTagLoader.add_multi_constructor("tag:", IgnoreTagLoader.construct_with_tag)

    SplitTagLoader.add_multi_constructor("!", SplitTagLoader.construct_with_tag)
    SplitTagLoader.add_multi_constructor("tag:", SplitTagLoader.construct_with_tag)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from torchwrench.core.make import (  # noqa: F401
    CUDA_IF_AVAILABLE,
    DeviceLike,
    DTypeLike,
    GeneratorLike,
    as_device,
    as_dtype,
    as_generator,
    get_default_device,
    get_default_dtype,
    get_default_generator,
    set_default_dtype,
    set_default_generator,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from abc import abstractmethod
from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    Iterator,
    Optional,
    TypeVar,
    Union,
)

from pythonwrench.collections import is_sorted
from pythonwrench.typing.classes import (
    SupportsGetitemIterLen,
    SupportsGetitemLen,
    SupportsIterLen,
)
from torch.utils.data.dataset import Dataset, IterableDataset
from torch.utils.data.dataset import IterableDataset as TorchIterableDataset
from torch.utils.data.dataset import Subset as TorchSubset

from torchwrench.types.tensor_subclasses import LongTensor1D

T = TypeVar("T", covariant=True)
U = TypeVar("U", covariant=True)

SizedDatasetLike = SupportsGetitemLen

T_Dataset = TypeVar("T_Dataset", bound=Dataset)
T_SizedDatasetLike = TypeVar("T_SizedDatasetLike", bound=SupportsGetitemLen)
T_SupportsIterLenDataset = TypeVar(
    "T_SupportsIterLenDataset",
    bound=SupportsGetitemIterLen,
)


class EmptyDataset(Dataset[None]):
    """Dataset placeholder. Raises StopIteration if __getitem__ is called."""

    def __getitem__(self, idx, /) -> None:  # type: ignore
        raise StopIteration

    def __len__(self) -> int:
        return 0


class _WrapperBase(Generic[T], Dataset[T]):
    def __init__(self, dataset: Any) -> None:
        Dataset.__init__(self)
        self.dataset = dataset

    @abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError

    def unwrap(self, recursive: bool = True) -> Union[SupportsGetitemLen, Dataset]:
        dataset = self.dataset
        continue_ = recursive and isinstance(
            dataset, (_WrapperBase, TorchSubset, TorchIterableDataset)
        )
        while continue_:
            dataset = dataset.dataset  # type: ignore
            continue_ = isinstance(dataset, _WrapperBase)
        return dataset  # type: ignore

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self.dataset)})"


class Wrapper(Generic[T], _WrapperBase[T]):
    @abstractmethod
    def __getitem__(self, idx, /) -> T:  # type: ignore
        raise NotImplementedError


class IterableWrapper(Generic[T], IterableDataset[T], _WrapperBase[T]):
    def __init__(
        self, dataset: Union[SupportsGetitemLen[T], SupportsIterLen[T]]
    ) -> None:
        IterableDataset.__init__(self)
        _WrapperBase.__init__(self, dataset)
        self.dataset: Union[SupportsGetitemLen[T], SupportsIterLen[T]]

    @abstractmethod
    def __iter__(self) -> Iterator[T]:
        raise NotImplementedError

    def _get_dataset_iter(self) -> Iterator[T]:
        if hasattr(self.dataset, "__iter__"):
            it = iter(self.dataset)
        else:
            it = (self.dataset[i] for i in range(len(self.dataset)))  # type: ignore
        return it


class TransformWrapper(Generic[T, U], Wrapper[T]):
    def __init__(
        self,
        dataset: SupportsGetitemLen[T],
        transform: Optional[Callable[[T], U]],
        condition: Optional[Callable[[T, int], bool]] = None,
    ) -> None:
        super().__init__(dataset)
        self._transform = transform
        self._condition = condition
        self.dataset: SupportsGetitemLen[T]

    def __getitem__(self, idx) -> Union[T, U]:  # type: ignore
        assert isinstance(idx, int)
        item = self.dataset[idx]
        if self._transform is not None and (
            self._condition is None or self._condition(item, idx)
        ):
            item = self._transform(item)
        return item

    def __len__(self) -> int:
        return len(self.dataset)

    @property
    def transform(self) -> Optional[Callable[[T], U]]:
        return self._transform

    @property
    def condition(self) -> Optional[Callable[[T, int], bool]]:
        return self._condition


class IterableTransformWrapper(IterableWrapper[T], Generic[T, U]):
    def __init__(
        self,
        dataset: Union[SupportsGetitemLen[T], SupportsIterLen[T]],
        transform: Optional[Callable[[T], U]],
        condition: Optional[Callable[[T, int], bool]] = None,
    ) -> None:
        super().__init__(dataset)
        self._transform = transform
        self._condition = condition

    def __iter__(self) -> Iterator[Union[T, U]]:  # type: ignore
        it = super()._get_dataset_iter()
        for i, item in enumerate(it):
            if self._transform is not None and (
                self._condition is None or self._condition(item, i)
            ):
                item = self._transform(item)
            yield item
        return

    def __len__(self) -> int:
        return len(self.dataset)

    @property
    def transform(self) -> Optional[Callable[[T], U]]:
        return self._transform

    @property
    def condition(self) -> Optional[Callable[[T, int], bool]]:
        return self._condition


class IterableSubset(IterableWrapper[T], Generic[T]):
    def __init__(
        self,
        dataset: Union[SupportsGetitemLen[T], SupportsIterLen[T]],
        indices: Union[Iterable[int], LongTensor1D],
    ) -> None:
        if isinstance(indices, LongTensor1D):
            indices = indices.tolist()
        else:
            indices = list(indices)

        if not all(idx >= 0 for idx in indices) or not is_sorted(indices):
            msg = f"Invalid argument {indices=}. (expected a sorted list of positive integers)"
            raise ValueError(msg)

        super().__init__(dataset)
        self._indices = indices

    def __iter__(self) -> Iterator[T]:
        it = super()._get_dataset_iter()

        cur_idx = 0
        item = next(it)

        for idx in self._indices:
            if cur_idx == idx:
                yield item
                continue

            while cur_idx < idx:
                cur_idx += 1
                item = next(it)

            yield item
        return

    def __len__(self) -> int:
        return len(self._indices)


class Subset(Generic[T], TorchSubset[T], Wrapper[T]):
    def __init__(self, dataset: SizedDatasetLike[T], indices: Iterable[int]) -> None:
        indices = list(indices)
        TorchSubset.__init__(self, dataset, indices)  # type: ignore
        Wrapper.__init__(self, dataset)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import itertools
import math
from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
    overload,
)

import torch
from pythonwrench.collections import all_eq as builtin_all_eq
from pythonwrench.collections import prod as builtin_prod
from pythonwrench.collections import unzip
from pythonwrench.functools import function_alias, identity
from pythonwrench.semver import Version
from pythonwrench.typing import BuiltinNumber, SupportsIterLen, T_BuiltinNumber
from pythonwrench.warnings import warn_once
from torch import Tensor, nn

from torchwrench.extras.numpy import np
from torchwrench.extras.pandas import pd
from torchwrench.nn import functional as F
from torchwrench.types._typing import (
    LongTensor,
    ScalarLike,
    T_TensorOrArray,
    TensorOrArray,
)
from torchwrench.types.guards import is_scalar_like
from torchwrench.types.tensor_subclasses import Tensor0D, Tensor1D, Tensor2D, Tensor3D
from torchwrench.utils import return_types

T = TypeVar("T", covariant=True)
U = TypeVar("U", covariant=True)


def count_parameters(
    model: nn.Module,
    *,
    recurse: bool = True,
    only_trainable: bool = False,
    buffers: bool = False,
) -> int:
    """Returns the number of parameters in a module."""
    params = (
        param
        for param in model.parameters(recurse)
        if not only_trainable or param.requires_grad
    )

    if buffers:
        params = itertools.chain(params, (buffer for buffer in model.buffers(recurse)))

    num_params = sum(param.numel() for param in params)
    return num_params


def find(
    value: Any,
    x: Tensor,
    *,
    default: Union[None, Tensor, BuiltinNumber] = None,
    dim: int = -1,
) -> LongTensor:
    """Return the index of the first occurrence of value in a tensor."""
    if x.ndim == 0:
        msg = f"Function 'find' does not supports 0-d tensors. (found {x.ndim=})"
        raise ValueError(msg)

    mask = x.eq(value)
    contains = mask.any(dim=dim)
    indices = mask.long().argmax(dim=dim)

    if default is None:
        if not contains.all():
            raise RuntimeError(f"Cannot find {value=} in tensor.")
        return indices  # type: ignore
    else:
        output = torch.where(contains, indices, default)
        return output  # type: ignore


@overload
def get_ndim(
    x: Union[ScalarLike, Tensor, np.ndarray, Iterable],
    *,
    use_first_for_list_tuple: bool = False,
    return_indicator: Literal[False] = False,
    return_default_on_invalid: Literal[False] = False,
    default: Any = -1,
    return_valid: Optional[bool] = None,
) -> int: ...


@overload
def get_ndim(
    x: Union[ScalarLike, Tensor, np.ndarray, Iterable],
    *,
    use_first_for_list_tuple: bool = False,
    return_indicator: Literal[False] = False,
    return_default_on_invalid: bool,
    default: U = -1,
    return_valid: Optional[bool] = None,
) -> Union[int, U]: ...


@overload
def get_ndim(
    x: Union[ScalarLike, Tensor, np.ndarray, Iterable],
    *,
    use_first_for_list_tuple: bool = False,
    return_indicator: Literal[True],
    return_default_on_invalid: Literal[False] = False,
    default: Any = -1,
    return_valid: Optional[bool] = None,
) -> return_types.ndim[int]: ...


@overload
def get_ndim(
    x: Union[ScalarLike, Tensor, np.ndarray, Iterable],
    *,
    use_first_for_list_tuple: bool = False,
    return_indicator: Literal[True],
    return_default_on_invalid: bool,
    default: U = -1,
    return_valid: Optional[bool] = None,
) -> return_types.ndim[Union[int, U]]: ...


def get_ndim(
    x: Union[ScalarLike, Tensor, np.ndarray, Iterable],
    *,
    use_first_for_list_tuple: bool = False,
    return_indicator: bool = False,
    return_default_on_invalid: bool = False,
    default: U = -1,
    return_valid: Optional[bool] = None,
) -> Union[Union[int, U], return_types.ndim[Union[int, U]]]:
    """Scan first argument to return its number of dimension(s). Works recursively with Tensors, numpy arrays and builtins types instances.

    Note: Sets and dicts are considered as scalars with a ndim equal to 0.

    Args:
        x: Input value to scan.
        use_first_for_list_tuple: If True, use first value to determine ndim for list and tuple argument. Otherwise it will scan each value in argument to determine its shape. defaults to False.
        return_indicator: If True, returns a tuple containing a boolean indicator if the data has an homogeneous ndim instead of raising a ValueError. defaults to False.
        return_default_on_invalid: If True and return_indicator=False, returns the default value instead of raising a ValueError. defaults to False.
        default: Value to return if input is a heterogeneous list/tuple. defaults to ().
        return_valid: Deprecated. Use return_indicator instead.

    Raises:
        ValueError if input has an heterogeneous number of dimensions and return_valid=False.
        TypeError if input has an unsupported type.
    """
    if return_valid is not None:
        msg = f"Deprecated argument {return_valid=}. Use return_indicator instead."
        warn_once(msg)
        return_indicator = return_valid
    del return_valid

    def _impl(
        x: Union[ScalarLike, Tensor, np.ndarray, Iterable],
    ) -> Tuple[bool, Union[int, U]]:
        if is_scalar_like(x):
            return True, 0
        elif isinstance(x, (Tensor, np.ndarray, np.generic, pd.DataFrame)):
            return True, x.ndim
        elif isinstance(x, (set, frozenset, dict)):
            return True, 0
        elif isinstance(x, (list, tuple)):
            valids_and_ndims = unzip(_impl(xi) for xi in x)
            if len(valids_and_ndims) == 0:
                return True, 1

            valids, ndims = valids_and_ndims
            if (use_first_for_list_tuple and valids[0]) or (
                all(valids) and builtin_all_eq(ndims)
            ):
                return True, ndims[0] + 1  # type: ignore
            else:
                return False, default
        else:
            raise TypeError(f"Invalid argument type {type(x)}.")

    valid, ndim = _impl(x)

    if not valid and not return_indicator and not return_default_on_invalid:
        msg = f"Invalid argument {x}. (cannot compute ndim for heterogeneous data)"
        raise ValueError(msg)

    if return_indicator:
        ndim = return_types.shape(valid, ndim)

    return ndim  # type: ignore


@function_alias(get_ndim)
def ndim(*args, **kwargs): ...


@overload
def get_shape(
    x: Union[
        ScalarLike, Tensor, np.ndarray, pd.DataFrame, list, tuple, set, frozenset, dict
    ],
    *,
    output_type: Callable[[Tuple[int, ...]], T] = identity,
    use_first_for_list_tuple: bool = False,
    return_indicator: Literal[False] = False,
    return_default_on_invalid: Literal[False] = False,
    default: Any = (),
    return_valid: Optional[bool] = None,
) -> T: ...


@overload
def get_shape(
    x: Union[
        ScalarLike, Tensor, np.ndarray, pd.DataFrame, list, tuple, set, frozenset, dict
    ],
    *,
    output_type: Callable[[Tuple[int, ...]], T] = identity,
    use_first_for_list_tuple: bool = False,
    return_indicator: Literal[False] = False,
    return_default_on_invalid: bool,
    default: U = (),
    return_valid: Optional[bool] = None,
) -> Union[T, U]: ...


@overload
def get_shape(
    x: Union[
        ScalarLike, Tensor, np.ndarray, pd.DataFrame, list, tuple, set, frozenset, dict
    ],
    *,
    output_type: Callable[[Tuple[int, ...]], T] = identity,
    use_first_for_list_tuple: bool = False,
    return_indicator: Literal[True],
    return_default_on_invalid: Literal[False] = False,
    default: Any = (),
    return_valid: Optional[bool] = None,
) -> return_types.shape[T]: ...


@overload
def get_shape(
    x: Union[
        ScalarLike, Tensor, np.ndarray, pd.DataFrame, list, tuple, set, frozenset, dict
    ],
    *,
    output_type: Callable[[Tuple[int, ...]], T] = identity,
    use_first_for_list_tuple: bool = False,
    return_indicator: Literal[True],
    return_default_on_invalid: bool,
    default: U = (),
    return_valid: Optional[bool] = None,
) -> return_types.shape[Union[T, U]]: ...


def get_shape(
    x: Union[
        ScalarLike, Tensor, np.ndarray, pd.DataFrame, list, tuple, set, frozenset, dict
    ],
    *,
    output_type: Callable[[Tuple[int, ...]], T] = identity,
    use_first_for_list_tuple: bool = False,
    return_indicator: bool = False,
    return_default_on_invalid: bool = False,
    default: U = (),
    return_valid: Optional[bool] = None,
) -> Union[T, U, return_types.shape[Union[T, U]]]:
    """Scan first argument to return its shape. Works recursively with Tensors, numpy arrays and builtins types instances.

    Note: Sets and dicts are considered as scalars with a shape equal to ().

    Args:
        x: Input value to scan.
        output_type: Output shape type. defaults to identity, which returns a tuple of ints.
        use_first_for_list_tuple: If True, use first value in sequences to determine shape for list and tuple argument. Otherwise it will scan each value in argument to determine its shape. defaults to False.
        return_indicator: If True, returns a tuple containing a boolean indicator if the data has an homogeneous shape instead of raising a ValueError. defaults to False.
        return_default_on_invalid: If True and return_indicator=False, returns the default value instead of raising a ValueError. defaults to False.
        default: Value to return if input is a heterogeneous list/tuple. This default value is NOT passed to the output_type() callable argument. defaults to ().
        return_valid: Deprecated. Use return_indicator instead.

    Raises:
        ValueError: if input has an heterogeneous shape and return_valid=False.
        TypeError: if input has an unsupported type.
    """
    if return_valid is not None:
        msg = f"Deprecated argument {return_valid=}. Use return_indicator instead."
        warn_once(msg)
        return_indicator = return_valid
    del return_valid

    def _impl(
        x: Union[ScalarLike, Tensor, np.ndarray, Iterable],
    ) -> Tuple[bool, Union[Tuple[int, ...], U]]:
        if is_scalar_like(x):
            return True, ()
        elif isinstance(x, (Tensor, np.ndarray, np.generic, pd.DataFrame)):
            return True, tuple(x.shape)
        elif isinstance(x, (set, frozenset, dict)):
            return True, ()
        elif isinstance(x, (list, tuple)):
            valids_and_shapes = unzip(_impl(xi) for xi in x)  # type: ignore
            if len(valids_and_shapes) == 0:
                return True, (0,)

            valids, shapes = valids_and_shapes
            if (use_first_for_list_tuple and valids[0]) or (
                all(valids) and builtin_all_eq(shapes)
            ):
                return True, (len(shapes),) + shapes[0]  # type: ignore
            else:
                return False, default
        else:
            raise TypeError(f"Invalid argument type {type(x)}.")

    valid, shape = _impl(x)

    if not valid and not return_indicator and not return_default_on_invalid:
        msg = f"Invalid argument {x}. (cannot compute shape for heterogeneous data)"
        raise ValueError(msg)

    if valid:
        shape = output_type(shape)  # type: ignore

    if return_indicator:
        shape = return_types.shape(valid, shape)

    return shape  # type: ignore


@function_alias(get_shape)
def shape(*args, **kwargs): ...


def ranks(x: Tensor, dim: int = -1, descending: bool = False) -> LongTensor:
    """Get the ranks of each value in range [0, x.shape[dim][."""
    return x.argsort(dim, descending).argsort(dim)  # type: ignore


def nelement(x: Union[ScalarLike, Tensor, np.ndarray, Iterable]) -> int:
    """Returns the number of elements in Tensor-like object."""
    if isinstance(x, Tensor):
        return x.nelement()
    elif isinstance(x, (np.ndarray, np.generic)):
        return x.size
    else:
        return builtin_prod(get_shape(x))


@overload
def prod(
    x: T_TensorOrArray,
    *,
    dim: Optional[int] = None,
    start: Any = 1,
) -> T_TensorOrArray: ...


@overload
def prod(
    x: Iterable[T_BuiltinNumber],
    *,
    dim: Any = None,
    start: T_BuiltinNumber = 1,
) -> T_BuiltinNumber: ...


def prod(
    x: Union[TensorOrArray, Iterable[T_BuiltinNumber]],
    *,
    dim: Optional[int] = None,
    start: T_BuiltinNumber = 1,
) -> Union[Tensor, T_BuiltinNumber]:
    """Returns the product of all elements in input."""
    if isinstance(x, Tensor):
        if dim is not None:
            return torch.prod(x, dim=dim) * start
        else:
            return torch.prod(x) * start
    elif isinstance(x, np.ndarray):
        return np.prod(x, axis=dim, initial=start)
    elif isinstance(x, Iterable):
        if dim is not None:
            msg = f"Invalid argument {dim=}. (expected None with {type(x)=})"
            raise ValueError(msg)
        return builtin_prod(x, start=start)  # type: ignore
    else:
        msg = (
            f"Invalid argument type {type(x)=}. (expected Tensor, ndarray or Iterable)"
        )
        raise TypeError(msg)


def average_power(
    x: T_TensorOrArray,
    dim: Union[int, Tuple[int, ...], None] = -1,
) -> T_TensorOrArray:
    """Compute average power of a signal along a specified dim/axis."""
    return (abs(x) ** 2).mean(dim)  # type: ignore


def mse(
    x1: Tensor,
    x2: Tensor,
    *,
    dim: Union[int, Tuple[int, ...], None] = None,
) -> Tensor:
    """Mean squared error function."""
    if dim is not None or Version(torch.__version__) >= "2.0.0":
        return ((x1 - x2) ** 2).mean(dim).sqrt()  # type: ignore
    else:
        return ((x1 - x2) ** 2).mean().sqrt()


def rmse(
    x1: Tensor,
    x2: Tensor,
    *,
    dim: Union[int, Tuple[int, ...], None] = None,
) -> Tensor:
    """Root mean squared error function."""
    return mse(x1, x2, dim=dim).sqrt()


def deep_equal(x: T, y: T, *args: T) -> bool:
    """Recursive comparison between objects.

    Supports Scalar-like, NDArrays, Tensors, DataFrames, Mapping and List-like objects.
    Unlike default equal, NaNs values are considered equal.
    Tensors and NDArrays of different shapes are supported and returns False.
    """
    others = (y,) + args
    for other in others:
        result = _deep_equal_binary(x, other)
        if not result:
            return False
    return True


def _deep_equal_binary(x: T, y: T) -> bool:
    if is_scalar_like(x) and is_scalar_like(y):
        x_isnan = math.isnan(x) if F.is_floating_point(x) else False
        y_isnan = math.isnan(y) if F.is_floating_point(y) else False
        return (x_isnan and y_isnan) or F.to_item(x == y)  # type: ignore

    if isinstance(x, Tensor) and isinstance(y, Tensor):
        x_isnan = x.isnan()
        y_isnan = y.isnan()
        return (
            (x.shape == y.shape)
            and bool((x_isnan == y_isnan).all().item())
            and torch.equal(x[~x_isnan], y[~y_isnan])
        )

    if isinstance(x, np.ndarray) and isinstance(y, np.ndarray):
        x_isnan = (
            np.isnan(x)
            if F.is_floating_point(x)
            else np.full(x.shape, False, dtype=bool)
        )
        y_isnan = (
            np.isnan(y)
            if F.is_floating_point(y)
            else np.full(y.shape, False, dtype=bool)
        )
        return (
            (x.shape == y.shape)
            and (x_isnan == y_isnan).all().item()
            and np.equal(x[~x_isnan], y[~y_isnan]).all().item()
        )

    if isinstance(x, pd.DataFrame) and isinstance(y, pd.DataFrame):
        if not (deep_equal(x.index, y.index) and deep_equal(x.columns, y.columns)):
            return False
        return (x.isna() == y.isna()).all(None).item() and (
            (x == y) | x.isna() | y.isna()
        ).all(None).item()  # type: ignore

    if isinstance(x, Mapping) and isinstance(y, Mapping):
        return deep_equal(list(x.items()), list(y.items()))

    if isinstance(x, SupportsIterLen) and isinstance(y, SupportsIterLen):
        return len(x) == len(y) and all(deep_equal(xi, yi) for xi, yi in zip(x, y))

    return (type(x) is type(y)) and (x == y)  # type: ignore


@overload
def stack(
    tensors: Union[List[Tensor0D], Tuple[Tensor0D, ...]],
    dim: int = 0,
    *,
    out: Optional[Tensor1D] = None,
) -> Tensor1D: ...


@overload
def stack(
    tensors: Union[List[Tensor1D], Tuple[Tensor1D, ...]],
    dim: int = 0,
    *,
    out: Optional[Tensor2D] = None,
) -> Tensor2D: ...


@overload
def stack(
    tensors: Union[List[Tensor2D], Tuple[Tensor2D, ...]],
    dim: int = 0,
    *,
    out: Optional[Tensor3D] = None,
) -> Tensor3D: ...


@overload
def stack(
    tensors: Union[List[Tensor], Tuple[Tensor, ...]],
    dim: int = 0,
    *,
    out: Optional[Tensor] = None,
) -> Tensor: ...


def stack(
    tensors,
    dim: int = 0,
    *,
    out: Optional[Tensor] = None,
) -> Tensor:  # type: ignore
    return torch.stack(tensors, dim=dim, out=out)


def cat(
    tensors: Union[List[Tensor], Tuple[Tensor, ...]],
    dim: int = 0,
    *,
    out: Optional[Tensor] = None,
) -> Tensor:
    return torch.cat(tensors, dim=dim, out=out)


@function_alias(cat)
def concat(*args, **kwargs): ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Literal, Optional, Union, overload

import torch
from torch.types import Device  # noqa: F401
from typing_extensions import TypeAlias

from torchwrench.core.dtype_enum import (
    DTypeEnum,
    enum_dtype_to_torch_dtype,
    str_to_torch_dtype,
)

DeviceLike: TypeAlias = Union[
    torch.device, None, Literal["default", "cuda_if_available"], str, int
]
DTypeLike: TypeAlias = Union[torch.dtype, None, Literal["default"], str, DTypeEnum]
GeneratorLike: TypeAlias = Union[torch.Generator, None, Literal["default"], int]

# Expose torch types
Generator = torch.Generator
device = torch.device


CUDA_IF_AVAILABLE = "cuda_if_available"


def get_default_device() -> torch.device:
    """Returns default device used when creating a tensor."""
    return torch.empty((0,)).device


def get_default_dtype() -> torch.dtype:
    return torch.get_default_dtype()


def get_default_generator() -> torch.Generator:
    return torch.default_generator


def set_default_dtype(dtype: DTypeLike) -> None:
    dtype = as_dtype(dtype)
    if dtype is not None:
        torch.set_default_dtype(dtype)


def set_default_generator(generator: GeneratorLike) -> None:
    generator = as_generator(generator)
    if generator is not None:
        torch.default_generator.set_state(generator.get_state())


@overload
def as_device(device: Literal[None]) -> None: ...


@overload
def as_device(
    device: Union[str, int, torch.device] = CUDA_IF_AVAILABLE,
) -> torch.device: ...


def as_device(device: DeviceLike = CUDA_IF_AVAILABLE) -> Optional[torch.device]:
    """Create concrete device object from device-like object."""
    if isinstance(device, (torch.device, type(None))) or device is ...:
        return device
    elif device == "default":
        return get_default_device()
    elif device == CUDA_IF_AVAILABLE:
        return torch.device("cuda" if torch.cuda.is_available() else "cpu")
    elif isinstance(device, (str, int)):
        return torch.device(device)
    else:
        msg = f"Invalid argument type {type(device)}. (expected torch.device, None, str, int or {CUDA_IF_AVAILABLE})"
        raise TypeError(msg)


@overload
def as_dtype(dtype: Literal[None] = None) -> None: ...


@overload
def as_dtype(dtype: Union[str, DTypeEnum, torch.dtype]) -> torch.dtype: ...


def as_dtype(dtype: DTypeLike = None) -> Optional[torch.dtype]:
    """Create concrete dtype object from dtype-like object."""
    if isinstance(dtype, (torch.dtype, type(None))) or dtype is ...:
        return dtype
    elif dtype == "default":
        return get_default_dtype()
    elif isinstance(dtype, DTypeEnum):
        return enum_dtype_to_torch_dtype(dtype)
    elif isinstance(dtype, str):
        return str_to_torch_dtype(dtype)
    else:
        msg = f"Invalid argument type {type(dtype)}. (expected torch.dtype, None, str or torchwrench.DTypeEnum)"
        raise TypeError(msg)


@overload
def as_generator(generator: Literal[None] = None) -> None: ...


@overload
def as_generator(
    generator: Union[int, torch.Generator, Literal["default"]],
) -> torch.Generator: ...


def as_generator(generator: GeneratorLike = None) -> Optional[torch.Generator]:
    """Create concrete generator object from generator-like object."""
    if isinstance(generator, (torch.Generator, type(None))) or generator is ...:
        return generator
    elif isinstance(generator, int):
        return torch.Generator().manual_seed(generator)
    elif generator == "default":
        return get_default_generator()
    else:
        msg = f"Invalid argument type {type(generator)}. (expected torch.Generator, None, int or 'default')"
        raise TypeError(msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import inspect
import unittest
from typing import Any, Callable, List, Tuple
from unittest import TestCase

import torch
from pythonwrench.inspect import get_fullname
from pythonwrench.typing import SupportsIterLen, isinstance_generic
from torch import Tensor

import torchwrench as tw


class TestFunctionsCompat(TestCase):
    def test_all_results(self) -> None:
        src_modules = [
            torch,
            torch.Tensor,
            torch.fft,
        ]
        all_base_fn_names = {
            src_module: dict.fromkeys(
                name
                for name in dir(src_module)
                if callable(getattr(src_module, name))
                and "_VariableFunctionsClass"
                not in get_fullname(getattr(src_module, name))
            )
            for src_module in src_modules
        }

        tested_fns = set()
        missed_fns = set()

        num_modules = 0
        targets: List[Tuple[Callable, Any, str]] = []

        for fn_name in dir(tw):
            tgt_fn = getattr(tw, fn_name)
            if not inspect.isfunction(tgt_fn) and not inspect.ismethod(tgt_fn):
                continue

            num_modules += 1

            fn_targets: List[Tuple[Callable, Any, str]] = []
            for src_module, fn_names in all_base_fn_names.items():
                if fn_name not in fn_names:
                    continue
                src_fn = getattr(src_module, fn_name)
                target = tgt_fn, src_fn, fn_name
                fn_targets.append(target)

            if len(fn_targets) == 0:
                missed_fns.add(fn_name)
            else:
                tested_fns.add(fn_name)

            targets += fn_targets

        seed = tw.randint(0, 10**6, ()).item()
        args_lst = [(torch.rand(10),)]
        for tgt_fn, src_fn, fn_name in targets:
            for args in args_lst:
                if tgt_fn == src_fn:
                    continue
                # skip some functions: function that have different signature (activations) and torch.save due to its non deterministic behaviour
                if fn_name in (
                    "log_softmax",
                    "softmax",
                    "sigmoid",
                    "log_sigmoid",
                    "save",
                ):
                    continue

                try:
                    tw.set_default_generator(seed)
                    result = tgt_fn(*args)
                except Exception as err:
                    result = err

                try:
                    tw.set_default_generator(seed)
                    expected = src_fn(*args)
                except Exception as err:
                    expected = err

                msg = f"Invalid results for '{fn_name}': {type(result)=} != {type(expected)=} from '{get_fullname(tgt_fn)}' and '{get_fullname(src_fn)}', with {args=}."
                if isinstance(result, Exception) and isinstance(expected, Exception):
                    pass
                elif isinstance(result, Tensor) and isinstance(expected, Tensor):
                    assert torch.equal(result, expected), msg
                elif isinstance_generic(
                    result, SupportsIterLen[Tensor]
                ) and isinstance_generic(expected, SupportsIterLen[Tensor]):
                    assert len(result) == len(expected), msg
                    for result_i, expected_i in zip(result, expected):
                        assert torch.equal(result_i, expected_i), msg
                else:
                    assert result == expected, msg

        print(
            f"Total base coverage hit: {len(tested_fns)}/{len(tested_fns) + len(missed_fns)}"
        )
        print(f"Hit_: {tested_fns}")
        print(f"Miss: {missed_fns}")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import itertools
import math
import struct
from typing import Callable, Union

import torch
from pythonwrench.checksum import (
    _cached_checksum_str,
    _checksum_iterable,
    checksum_any,  # noqa: F401
    checksum_bytes,
    checksum_dict,
    checksum_float,
    checksum_list_tuple,
    checksum_str,
    register_checksum_fn,
)
from pythonwrench.inspect import get_fullname
from torch import Tensor, nn

from torchwrench.core.packaging import _NUMPY_AVAILABLE, _PANDAS_AVAILABLE
from torchwrench.extras.numpy import np
from torchwrench.extras.pandas import pd
from torchwrench.nn.functional.predicate import is_complex, is_floating_point


@register_checksum_fn(pd.DataFrame)
def checksum_dataframe(x: pd.DataFrame, **kwargs) -> int:
    if not _PANDAS_AVAILABLE:
        msg = "Cannot call function 'checksum_dataframe' because optional dependency 'pandas' is not installed. Please install it using 'pip install torchwrench[extras]'"
        raise NotImplementedError(msg)

    kwargs["accumulator"] = kwargs.get("accumulator", 0) + _cached_checksum_str(
        get_fullname(x)
    )
    xdict = x.to_dict()
    return checksum_dict(xdict, **kwargs)  # type: ignore


@register_checksum_fn(pd.Series)
def checksum_series(x: pd.Series, **kwargs) -> int:
    if not _PANDAS_AVAILABLE:
        msg = "Cannot call function 'checksum_series' because optional dependency 'pandas' is not installed. Please install it using 'pip install torchwrench[extras]'"
        raise NotImplementedError(msg)

    kwargs["accumulator"] = kwargs.get("accumulator", 0) + _cached_checksum_str(
        get_fullname(x)
    )
    xlist = x.tolist()
    return checksum_list_tuple(xlist, **kwargs)  # type: ignore


@register_checksum_fn((torch.dtype, np.dtype))
def checksum_dtype(x: Union[torch.dtype, np.dtype], **kwargs) -> int:
    kwargs["accumulator"] = kwargs.get("accumulator", 0) + _cached_checksum_str(
        get_fullname(x)
    )
    xstr = str(x)
    return checksum_str(xstr, **kwargs)


@register_checksum_fn(nn.Module)
def checksum_module(
    x: nn.Module,
    *,
    only_trainable: bool = False,
    with_names: bool = False,
    buffers: bool = False,
    training: bool = False,
    **kwargs,
) -> int:
    """Compute a simple checksum over module parameters."""
    training = x.training
    x.train(training)

    if with_names:
        params_it = (
            (n, p)
            for n, p in x.named_parameters()
            if not only_trainable or p.requires_grad  # type: ignore
        )
    else:
        params_it = (
            param
            for param in x.parameters()
            if not only_trainable or param.requires_grad
        )

    if not buffers:
        iterator = params_it
    elif with_names:
        buffers_it = (name_buffer for name_buffer in x.named_buffers())
        iterator = itertools.chain(params_it, buffers_it)
    else:
        buffers_it = (buffer for buffer in x.buffers())
        iterator = itertools.chain(params_it, buffers_it)

    csum = _checksum_iterable(iterator, **kwargs)
    x.train(training)
    return csum


# Intermediate functions
@torch.inference_mode()
@register_checksum_fn(Tensor)
def checksum_tensor(x: Tensor, **kwargs) -> int:
    """Compute a simple checksum of a tensor. Order of values matter for the checksum."""
    return _checksum_tensor_array_like(
        x,
        nan_to_num_fn=torch.nan_to_num,
        **kwargs,
    )


@torch.inference_mode()
@register_checksum_fn((np.ndarray, np.generic))
def checksum_numpy(x: Union[np.ndarray, np.generic], **kwargs) -> int:
    """Compute a simple checksum of a tensor. Order of values matter for the checksum."""
    return _checksum_tensor_array_like(
        x,
        nan_to_num_fn=np.nan_to_num,
        **kwargs,
    )


# Private functions
def _checksum_tensor_array_like(
    x: Union[Tensor, np.ndarray, np.generic],
    *,
    nan_to_num_fn: Callable,
    **kwargs,
) -> int:
    if is_floating_point(x) or is_complex(x):
        nan_csum = checksum_float(math.nan, **kwargs)
        neginf_csum = checksum_float(-math.inf, **kwargs)
        posinf_csum = checksum_float(math.inf, **kwargs)
        x = nan_to_num_fn(
            x,
            nan=nan_csum,
            neginf=neginf_csum,
            posinf=posinf_csum,
        )

    # Ensure that accumulator exists
    kwargs["accumulator"] = kwargs.get("accumulator", 0)

    kwargs["accumulator"] += checksum_dtype(x.dtype, **kwargs)
    kwargs["accumulator"] += _checksum_iterable(x.shape, **kwargs)
    kwargs["accumulator"] += _cached_checksum_str(get_fullname(x))

    if isinstance(x, (np.ndarray, np.generic)):
        xbytes = x.tobytes()
        csum = checksum_bytes(xbytes, **kwargs)
    elif isinstance(x, Tensor):
        if _NUMPY_AVAILABLE:
            xbytes = x.detach().cpu().numpy().tobytes()
        else:
            xbytes = _serialize_tensor_to_bytes(x)
        csum = checksum_bytes(xbytes, **kwargs)
    else:
        msg = f"Invalid argument type {type(x)}. (expected ndarray or Tensor)"
        raise TypeError(msg)

    return csum


def _serialize_tensor_to_bytes(x: Tensor) -> bytes:
    """Convert tensor data to bytes, but very slow compare to numpy' tobytes() method."""
    x = x.view(torch.int8).view(-1)
    xbytes = struct.pack(f"{len(x)}b", *x)
    return xbytes



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Functional interface."""

from torch.nn.functional import *  # type: ignore

from .activation import log_softmax_multidim, softmax_multidim
from .checksum import checksum_any
from .cropping import crop_dim, crop_dims
from .indices import (
    get_inverse_perm,
    get_perm_indices,
    insert_at_indices,
    randperm_diff,
    remove_at_indices,
)
from .make import (
    as_device,
    as_dtype,
    as_generator,
    get_default_device,
    get_default_dtype,
    get_default_generator,
    set_default_dtype,
    set_default_generator,
)
from .mask import (
    generate_square_subsequent_mask,
    lengths_to_non_pad_mask,
    lengths_to_pad_mask,
    lengths_to_ratios,
    masked_equal,
    masked_mean,
    masked_sum,
    non_pad_mask_to_lengths,
    non_pad_mask_to_ratios,
    pad_mask_to_lengths,
    pad_mask_to_ratios,
    ratios_to_lengths,
    ratios_to_non_pad_mask,
    ratios_to_pad_mask,
    tensor_to_lengths,
    tensor_to_non_pad_mask,
    tensor_to_pad_mask,
    tensor_to_tensors_list,
    tensors_list_to_lengths,
)
from .multiclass import (
    index_to_name,
    index_to_onehot,
    name_to_index,
    name_to_onehot,
    one_hot,
    onehot_to_index,
    onehot_to_name,
    probs_to_index,
    probs_to_name,
    probs_to_onehot,
)
from .multilabel import (
    indices_to_multihot,
    indices_to_multinames,
    multi_indices_to_multihot,
    multi_indices_to_multinames,
    multihot_to_indices,
    multihot_to_multi_indices,
    multihot_to_multinames,
    multinames_to_indices,
    multinames_to_multi_indices,
    multinames_to_multihot,
    probs_to_indices,
    probs_to_multi_indices,
    probs_to_multihot,
    probs_to_multinames,
)
from .new import arange, empty, full, ones, rand, randint, randperm, zeros
from .others import (
    average_power,
    cat,
    concat,
    count_parameters,
    deep_equal,
    find,
    get_ndim,
    get_shape,
    mse,
    ndim,
    nelement,
    prod,
    ranks,
    rmse,
    shape,
    stack,
)
from .padding import cat_padded_batch, pad_and_stack_rec, pad_dim, pad_dims
from .powerset import multilabel_to_powerset, powerset_to_multilabel
from .predicate import (
    all_eq,
    all_ne,
    is_complex,
    is_convertible_to_tensor,
    is_floating_point,
    is_full,
    is_sorted,
    is_stackable,
    is_unique,
)
from .segments import (
    activity_to_segments,
    activity_to_segments_list,
    segments_list_to_activity,
    segments_to_activity,
    segments_to_segments_list,
)
from .transform import (  # noqa: F811
    as_tensor,
    flatten,
    identity,
    move_to,
    move_to_rec,
    pad_and_crop_dim,
    recursive_to,
    repeat_interleave_nd,
    resample_nearest_freqs,
    resample_nearest_rates,
    resample_nearest_steps,
    shuffled,
    squeeze,
    squeeze_,
    squeeze_copy,
    to_item,
    to_tensor,
    top_k,
    top_p,
    topk,
    transform_drop,
    unsqueeze,
    unsqueeze_,
    unsqueeze_copy,
    view_as_complex,
    view_as_real,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Any, Iterable, List, Sized, Tuple, TypeVar, Union, overload

import torch
from pythonwrench.collections import all_eq as builtin_all_eq
from pythonwrench.collections import all_ne as builtin_all_ne
from pythonwrench.collections import is_sorted as builtin_is_sorted
from pythonwrench.functools import function_alias
from pythonwrench.typing import is_builtin_number, isinstance_generic
from torch import Tensor
from typing_extensions import TypeGuard

from torchwrench.extras.numpy import (
    ACCEPTED_NUMPY_DTYPES,
    np,
    numpy_all_eq,
    numpy_all_ne,
    numpy_is_complex,
    numpy_is_floating_point,
)
from torchwrench.nn.functional.others import nelement
from torchwrench.types._typing import (
    ComplexFloatingTensor,
    FloatingTensor,
    ScalarLike,
    T_TensorOrArray,
    Tensor0D,
    TensorOrArray,
)
from torchwrench.types.guards import is_scalar_like

T = TypeVar("T")
U = TypeVar("U")


def is_stackable(
    tensors: Union[List[Any], Tuple[Any, ...]],
) -> TypeGuard[Union[List[Tensor], Tuple[Tensor, ...]]]:
    """Returns True if inputs can be passed to `torch.stack` function, i.e. contains a non-empty list or tuple of tensors with the same shape."""
    if not isinstance_generic(tensors, List[Tensor]) and not isinstance_generic(
        tensors, Tuple[Tensor, ...]
    ):
        return False
    if len(tensors) == 0:
        return False
    shape0 = tensors[0].shape
    result = all(tensor.shape == shape0 for tensor in tensors[1:])
    return result


def is_convertible_to_tensor(x: Any) -> bool:
    """Returns True if inputs can be passed to `torch.as_tensor` function.

    This function returns False for heterogeneous inputs like `[[], 1]`, but this kind of value can be accepted by `torch.as_tensor`.
    """
    if isinstance(x, Tensor):
        return True
    else:
        return __can_be_converted_to_tensor_nested(x)


def __can_be_converted_to_tensor_list_tuple(x: Union[List, Tuple]) -> bool:
    if len(x) == 0:
        return True

    valid_items = all(__can_be_converted_to_tensor_nested(xi) for xi in x)
    if not valid_items:
        return False

    # If all values are scalars-like items
    if all((not isinstance(xi, Sized) or isinstance(xi, Tensor0D)) for xi in x):
        return True

    # If all values are sized items with same size
    elif all(isinstance(xi, Sized) for xi in x):
        len0 = len(x[0])
        return all(len(xi) == len0 for xi in x[1:])

    else:
        return False


def __can_be_converted_to_tensor_nested(
    x: Any,
) -> bool:
    if is_builtin_number(x):
        return True
    elif isinstance(x, Tensor0D):
        return True
    elif isinstance(x, (np.ndarray, np.generic)) and x.dtype in ACCEPTED_NUMPY_DTYPES:
        return True
    elif isinstance(x, (List, Tuple)):
        return __can_be_converted_to_tensor_list_tuple(x)
    else:
        return False


@overload
def is_floating_point(x: Tensor) -> TypeGuard[FloatingTensor]: ...


@overload
def is_floating_point(x: np.ndarray) -> TypeGuard[np.ndarray]: ...


@overload
def is_floating_point(x: float) -> TypeGuard[float]: ...


@overload
def is_floating_point(
    x: Any,
) -> TypeGuard[Union[FloatingTensor, np.ndarray, float]]: ...


def is_floating_point(x: Any) -> TypeGuard[Union[FloatingTensor, np.ndarray, float]]:
    """Returns True if object is a/contains floating-point object(s)."""
    if isinstance(x, Tensor):
        return x.is_floating_point()
    elif isinstance(x, (np.ndarray, np.generic)):
        return numpy_is_floating_point(x)
    else:
        return isinstance(x, float)


@overload
def is_complex(x: Tensor) -> TypeGuard[ComplexFloatingTensor]: ...


@overload
def is_complex(x: np.ndarray) -> TypeGuard[np.ndarray]: ...


@overload
def is_complex(x: complex) -> TypeGuard[complex]: ...


@overload
def is_complex(
    x: Any,
) -> TypeGuard[Union[ComplexFloatingTensor, np.ndarray, complex]]: ...


def is_complex(x: Any) -> TypeGuard[Union[ComplexFloatingTensor, np.ndarray, complex]]:
    """Returns True if object is a/contains complex-valued object(s)."""
    if isinstance(x, Tensor):
        return x.is_complex()
    elif isinstance(x, (np.ndarray, np.generic)):
        return numpy_is_complex(x)
    else:
        return isinstance(x, complex)


def is_sorted(
    x: Union[Tensor, np.ndarray, Iterable],
    *,
    reverse: bool = False,
    strict: bool = False,
) -> bool:
    """Returns True if the sequence is sorted."""
    if isinstance(x, (Tensor, np.ndarray)):
        if x.ndim != 1:
            msg = f"Invalid number of dims in argument {x.ndim=}. (expected 1)"
            raise ValueError(msg)

        prev = slice(None, -1)
        next_ = slice(1, None)

        if not reverse and not strict:
            result = x[prev] <= x[next_]
        elif not reverse and strict:
            result = x[prev] < x[next_]
        elif reverse and not strict:
            result = x[prev] >= x[next_]
        else:  # reverse and strict
            result = x[prev] > x[next_]

        result = result.all().item()
        return result  # type: ignore

    elif isinstance(x, Iterable):
        return builtin_is_sorted(x, reverse=reverse, strict=strict)

    else:
        raise TypeError(f"Invalid argument type {type(x)=}.")


@overload
def all_eq(
    x: Union[Tensor, np.ndarray, ScalarLike, Iterable],
    dim: None = None,
) -> bool: ...


@overload
def all_eq(
    x: T_TensorOrArray,
    dim: int,
) -> T_TensorOrArray: ...


def all_eq(
    x: Union[T_TensorOrArray, ScalarLike, Iterable],
    dim: Union[int, None] = None,
) -> Union[bool, T_TensorOrArray]:
    """Check if all elements are equal in a tensor, ndarray, iterable or scalar object."""
    if isinstance(x, Tensor):
        if dim is None:
            if x.ndim == 0 or x.nelement() == 0:
                return True
            x = x.reshape(-1)
            return (x[0] == x[1:]).all().item()  # type: ignore
        else:
            slices: List[Union[slice, int, None]] = [slice(None) for _ in range(x.ndim)]
            slices[dim] = 0
            slices.insert(dim + 1, None)
            return (x == x[tuple(slices)]).all(dim)  # type: ignore

    elif isinstance(x, (np.ndarray, np.generic)):
        return numpy_all_eq(x, dim=dim)  # type: ignore

    elif dim is not None:
        raise ValueError(f"Invalid argument {dim=} with {type(x)=}.")

    elif is_scalar_like(x):
        return True

    elif isinstance(x, Iterable):
        return builtin_all_eq(x)

    else:
        raise TypeError(f"Invalid argument type {type(x)=}.")


def all_ne(x: Union[Tensor, np.ndarray, ScalarLike, Iterable]) -> bool:
    """Check if all elements are NOT equal in a tensor, ndarray, iterable or scalar object."""
    if isinstance(x, Tensor):
        return len(torch.unique(x)) == x.nelement()
    elif isinstance(x, (np.ndarray, np.generic)):
        return numpy_all_ne(x)
    elif is_scalar_like(x):
        return True
    elif isinstance(x, Iterable):
        return builtin_all_ne(x)
    else:
        raise TypeError(f"Invalid argument type {type(x)=}.")


def is_full(x: TensorOrArray, target: Any = ...) -> bool:
    """Check if all element are equal to target in a tensor or array. Accept an optional value 'target' to specified the expected value."""
    if nelement(x) == 0 and target is not ...:
        return False
    if nelement(x) == 0 and target is ...:
        return True

    indices = tuple([0] * x.ndim)
    first_elem = x[indices]
    return (target is ... or first_elem == target) and all_eq(x)


@function_alias(all_ne)
def is_unique(*args, **kwargs): ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import unittest
from unittest import TestCase

import yaml
from yaml import FullLoader, SafeLoader
from yaml.constructor import ConstructorError

from torchwrench.core.packaging import _YAML_AVAILABLE
from torchwrench.extras.yaml import (
    IgnoreTagLoader,
    SplitTagLoader,
    dump_yaml,
    load_yaml,
)


class TestYaml(TestCase):
    def test_yaml_load_examples(self) -> None:
        if not _YAML_AVAILABLE:
            return None

        dumped = "a: !!python/tuple\n- 1\n- 2"

        result = yaml.load(dumped, Loader=IgnoreTagLoader)
        expected = {"a": [1, 2]}
        assert result == expected

        result = yaml.load(dumped, Loader=FullLoader)
        expected = {"a": (1, 2)}
        assert result == expected

        with self.assertRaises(ConstructorError):
            yaml.load(dumped, Loader=SafeLoader)

        result = yaml.load(dumped, Loader=SplitTagLoader)
        expected = {"a": {"_target_": "yaml.org,2002:python/tuple", "_args_": [1, 2]}}
        assert result == expected

        data = load_yaml(io.StringIO(dumped), Loader=FullLoader)
        assert dump_yaml(data) == f"{dumped}\n", f"{data=}"
        assert dump_yaml(data, to_builtins=True) == "a:\n- 1\n- 2\n", f"{data=}"


if __name__ == "__main__":
    unittest.main()
